#!/usr/bin/env python3
#
# Copyright 2019. Clumio, Inc.

import boto3
import click
import pandas as pd
from botocore.exceptions import ClientError
import openpyxl
import datetime

nl = '\n'

@click.group()
def cli():
    '''

    Command line tools by Clumio, inc.

    Maintained by Luke McCleary | Luke@clumio.com

    '''
    pass


@cli.command()
@click.option('-id', '--key-id', prompt=True)
@click.option('-sk', '--secret-key', prompt=True, hide_input=True)
@click.option('-o', '--outfile', default='Clumio_Calipers.xlsx')
@click.option('-rb', '--count-blocks', default='True')
@click.option('-g', '--geo', multiple=True, required=True, default=['am'],
              type=click.Choice(['eu', 'am', 'apac'], case_sensitive=True))
def report(key_id, secret_key, outfile, count_blocks, geo):
    '''
    Generates an excel report on EC2, EBS, RDS, etc assets. Defaults to
    North and South American regions, non-govcloud. Other regions,
    or for multiple regions, us the --geo or -g flags.

    Example 1: Let the script prompt for required inputs: \n
        Python3 -m calipers report

    Example 2: Specify output file name and a geo: \n
        Python3 -m calipers report -o filename.xlsx -g eu

    Example 3: Specify output file name and multiple geos: \n
        Python3 -m calipers report -o filename.xlsx -g eu -g am

    Americas = am \n
    Europe = eu \n
    Asia = apac \n

    '''
    global read_blocks
    read_blocks = count_blocks

    americas = ['us-east-1', 'us-east-2', 'us-west-1',
                'us-west-2', 'ca-central-1', 'sa-east-1']
    europe = ['eu-central-1', 'eu-west-1', 'me-south-1',
              'eu-west-2', 'eu-west-3', 'eu-north-1']
    asiapac = ['ap-east-1', 'ap-south-1',
               'ap-northeast-3', 'ap-northeast-2', 'ap-southeast-1',
               'ap-southeast-2', 'ap-northeast-1']

    regions = []

    if 'am' in geo:
        for i in americas:
            regions.append(i)
    if 'eu' in geo:
        for i in europe:
            regions.append(i)
    if 'apac' in geo:
        for i in asiapac:
            regions.append(i)

    print(f'Reporting on the following AWS Regions {regions}{nl}')

    '''initialize dictionaries'''
    ec2_master = []
    active_ebs_vols_master = []
    orphan_ebs_vols_master = []
    ebs_backed_snaps_master = []
    orphan_ebs_snaps_master = []
    rds_instances_master = []
    rds_clusters_master = []

    for region in regions:
        '''Regional iterative with connection to relevent service'''
        try:
            'Gather regional data and append to global/master dictionaries'

            '''********  EC2 / EBS  **********'''

            '''Create Connection Objects'''
            global ebs_conn
            global ec2_conn
            ec2_conn = connect_service(key_id, secret_key, region, 'ec2')
            ebs_conn = connect_service(key_id, secret_key, region, 'ebs')
            
            '''fetch data + build ec2 report dictionary'''
            for i in (ec2_report(region)):
                ec2_master.append(i)

            '''fetch data for EBS report dictionaries'''
            in_use_ebs_list, in_use_ebs_response = get_ebs_vols(
                region, 'in-use')
            avail_ebs_list, avail_ebs_response = get_ebs_vols(
                region, 'available')
            in_use_ebs_snap_reponse = get_ebs_snapshots(
                region, in_use_ebs_list)
            avail_ebs_snap_reponse = get_ebs_snapshots(
                region, avail_ebs_list)

            '''feed in data to generate EBS volume report dictionaries'''
            in_use_dict = ebs_vols_report(
                region, in_use_ebs_response, in_use_ebs_snap_reponse)
            avail_dict = ebs_vols_report(
                region, avail_ebs_response, avail_ebs_snap_reponse)
            for a in in_use_dict:
                active_ebs_vols_master.append(a)
            for o in avail_dict:
                orphan_ebs_vols_master.append(o)

            '''feed in data to generate EBS Snapshot report Dictionaries'''
            all_vols = avail_ebs_list + in_use_ebs_list
            all_ebs_backed_snaps = (in_use_ebs_snap_reponse +
                                    avail_ebs_snap_reponse)
            orphan_ebs_snaps = get_orphan_ebs_snaps(all_vols)

            for i in (ebs_snap_report(region, all_ebs_backed_snaps)):
                ebs_backed_snaps_master.append(i)
            for i in ebs_snap_report(region, orphan_ebs_snaps):
                orphan_ebs_snaps_master.append(i)

            '''RDS'''
            global rds_conn
            rds_conn = connect_service(key_id, secret_key, region, 'rds')
            for i in (rds_inst_report(region)):
                rds_instances_master.append(i)
            for i in (rds_clusters_report(region)):
                rds_clusters_master.append(i)

        except Exception as e:
            print('A general error occured returning data from AWS\n' + str(e))

    try:
        '''convert returned dicts to dataframes'''
        ec2_df = dataframer(
            ec2_master, 'InstanceId', 'EC2')
        active_ebs_df = dataframer(
            active_ebs_vols_master, 'VolumeId', 'EBS Vols')
        orph_ebs_df = dataframer(
            orphan_ebs_vols_master, 'VolumeId', 'EBS Vols')
        ebs_snap_df = dataframer(
            ebs_backed_snaps_master, 'SnapshotId', 'EBS Snaps')
        orph_ebs_snap_df = dataframer(
            orphan_ebs_snaps_master, 'SnapshotId', 'EBS Snaps')
        rds_inst_df = dataframer(
            rds_instances_master, 'DBI Identifier', 'RDS Instances')
        rds_clust_df = dataframer(
            rds_clusters_master, 'DBC Identifier', 'RDS Clusters')
    except Exception as e:
        print(f'Error adding python dictionary to pandas dataframe:{nl} {e}')

    print('''
    :::::::::::::::::::::::::::::::::::::::::::::\n
    AWS data collection complete. Tidying up... \n
    :::::::::::::::::::::::::::::::::::::::::::::\n
    ''')

    '''Analyze / aggregate the results'''

    '''EC2 / EBS'''
    try:
        ec2_num = len(ec2_df.index)
    except Exception:
        ec2_num = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        attached_ebs_num = len(active_ebs_df.index)
    except Exception:
        attached_ebs_num = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        attached_ebs_prov = active_ebs_df['Size (GB)'].sum()
    except Exception:
        attached_ebs_prov = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        orph_ebs_num = len(orph_ebs_df.index)
    except Exception:
        orph_ebs_num = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        orph_ebs_prov = orph_ebs_df['Size (GB)'].sum()
    except Exception:
        orph_ebs_prov = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        unencrypted_ebs = ((orph_ebs_df["Encrypted?"].str.count('No')).sum() +
                           (active_ebs_df["Encrypted?"].str.count('No')).sum())
    except Exception:
        unencrypted_ebs = None
        print('encountered summarizing error. Possible missing values')
        pass

    '''unprotected ebs'''
    try:
        unprotected_active = active_ebs_df["Total Snaps"].isin(['0'])
    except Exception:
        unprotected_active = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        unprotected_orph = orph_ebs_df["Total Snaps"].isin(['0'])
    except Exception:
        unprotected_orph = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        protected_ebs = (len(unprotected_active.loc[unprotected_active]) +
                         len(unprotected_orph.loc[unprotected_orph]))
    except Exception:
        protected_ebs = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        unprotected_ebs = (attached_ebs_num + orph_ebs_num - protected_ebs)
    except Exception:
        unprotected_ebs = None
        print('encountered summarizing error. Possible missing values')
        pass

    '''EBS Snaps'''
    try:
        backed_snaps_num = len(ebs_snap_df.index)
    except Exception:
        backed_snaps_num = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        orphaned_snaps_num = len(orph_ebs_snap_df.index)
    except Exception:
        orphaned_snaps_num = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        old_active = ebs_snap_df[ebs_snap_df['60+ Days Old']]
    except Exception:
        old_active = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        old_orph = orph_ebs_snap_df.loc[orph_ebs_snap_df['60+ Days Old']]
    except Exception:
        old_orph = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        ebs_snaps_60plus = (len(old_active) + len(old_orph))
    except Exception:
        ebs_snaps_60plus = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        active_ebs_snap_s3 = ebs_snap_df['Snapshot Size on S3 (GB)'].sum()
    except Exception:
        active_ebs_snap_s3 = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        orphan_ebs_snap_s3 = orph_ebs_snap_df['Snapshot Size on S3 (GB)'].sum()
    except Exception:
        orphan_ebs_snap_s3 = None
        print('encountered summarizing error. Possible missing values')
        pass

    '''RDS'''
    try:
        rds_clustersGB = rds_clust_df['AllocatedStorage (GB)'].sum()
    except Exception:
        rds_clustersGB = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        rds_instancesGB = rds_inst_df['DB Provisioned Capacity (GB)'].sum()
    except Exception:
        rds_instancesGB = None
        print('encountered summarizing error. Possible missing values')
        pass

    try:
        provisioned_RDS = rds_instancesGB + rds_clustersGB
    except Exception:
        provisioned_RDS = None
        print('encountered summarizing error. Possible missing values\n')
        pass

    summary = []
    summary.append({
        'EC2': '-',
        'EC2 Instances': ec2_num,
        '-': '-',
        'EBS Volumes': '-',
        'Attached EBS Volumes': attached_ebs_num,
        'Attached EBS Provisioned (GB)': attached_ebs_prov,
        'Available EBS Volumes': orph_ebs_num,
        'Available EBS Provisioned (GB)': orph_ebs_prov,
        'Unencrypted Volumes': unencrypted_ebs,
        '--': '--',
        'EBS Snapshots': '-',
        'Unprotected Volumes (no snaps)': unprotected_ebs,
        'EBS Backed Snaps': backed_snaps_num,
        'EBS Backed Snap space on S3 (GB)': active_ebs_snap_s3,
        'Orphaned EBS Snaps': orphaned_snaps_num,
        'Orphan EBS Snap space on S3 (GB)': orphan_ebs_snap_s3,
        'Unencrypted Volumes': unencrypted_ebs,
        'EBS Snaps > 60 Days Old': ebs_snaps_60plus,
        '----': '----',
        'RDS': ' ',
        'Total RDS Provisioned (GB)': provisioned_RDS
    })

    summary_df = pd.DataFrame(summary).transpose()
    summary_df.rename(columns={0: 'Value'}, inplace=True)

    try:
        '''pandas commands to merge reports into excel workbooks by tabs'''
        writer = pd.ExcelWriter(outfile)
        summary_df.to_excel(writer, 'AWS_Summary')
        ec2_df.to_excel(writer, 'Running_Stopped_EC2')
        active_ebs_df.to_excel(writer, 'In-Use EBS Volumes')
        orph_ebs_df.to_excel(writer, 'Available EBS Volumes')
        ebs_snap_df.to_excel(writer, 'EBS Backed Snaphots')
        orph_ebs_snap_df.to_excel(writer, 'Orphan EBS Snapshots')
        rds_inst_df.to_excel(writer, 'RDS_Instances')
        rds_clust_df.to_excel(writer, 'RDS_Clusters')
        writer.save()

    except Exception as e:
        print('\nError Writing to Excel File\n' + str(e))

    print(summary_df)


def ec2_report(region):
    '''initialize variables'''
    custom_ec2_dict = []
    response = []
    response_dict = []
    ec2_vols = []
    vols_dict = []

    try:
        '''fetch running and stopped ec2 instances with paginator'''
        print(f'{nl}Fetching EC2 Data from {region}...')

        pag = ec2_conn.get_paginator('describe_instances')
        for page in pag.paginate(Filters=[
                {
                    'Name': 'instance-state-name',
                    'Values': ['running', 'stopped']
                },

        ], PaginationConfig={
                'PageSize': 150
        }
        ):
            for i in page['Reservations']:
                response.append(i)
    except Exception as e:
        print(f'Failed to fetch EC2 Instances {e}')

    '''this line can probs be removed by adding [instances] to array above'''
    for d in range(len(response)):
        for i in range(len(response[d]['Instances'])):
            response_dict.append(response[d]['Instances'][i])

    print(f'{nl}Found {len(response_dict)} EC2 instances {region}')

    try:
        '''fetch all volumes associated with EC2 Instances'''
        ec2_vols = []
        for i in range(len(response_dict)):
            for v in range(len(response_dict[i]['BlockDeviceMappings'])):
                ec2_vols.append(response_dict[i]['BlockDeviceMappings'][v]
                                ['Ebs']['VolumeId'])

        '''fetch detailed volume info for associated EC2 instances'''
        p = Paginator(ec2_vols, 150)
        for i in p:
            response = ec2_conn.describe_volumes(Filters=[
                {
                    'Name': 'volume-id',
                    'Values': i,
                }
            ])['Volumes']
            for i in response:
                vols_dict.append(i)
    except Exception as e:
        print(f'Failed to fetch volumes assocated with EC2 instances {e}')

    try:
        '''iterate through regions appending data to datastructure'''
        for i in range(len(response_dict)):
            '''store volume IDs in local variable for loop'''
            vols = []
            tags = None
            for v in range(len(response_dict[i]['BlockDeviceMappings'])):
                vols.append(response_dict[i]['BlockDeviceMappings'][v]['Ebs']
                            ['VolumeId'])

            '''add up vol sizes'''
            ebs_agg = 0
            for v in vols:
                for d in range(len(vols_dict)):
                    if v == vols_dict[d]['VolumeId']:
                        ebs_agg = ebs_agg + vols_dict[d]['Size']

            '''collect tag info'''
            try:
                tags = response_dict[i]['Tags']
            except Exception:
                tags = None
                pass

            '''write ec2 data to data structure'''
            custom_ec2_dict.append({
                'InstanceId': response_dict[i]['InstanceId'],
                'Type': response_dict[i]['InstanceType'],
                'Region': str(region),
                'Power-State': response_dict[i]['State']['Name'],
                'Total EBS Provisioned (GB)': ebs_agg,
                'Attached-Volumes': str(vols),
                'Tags': tags
            })
    except Exception as e:
        print(f'No EC2 Instances found in {region}:{nl}{e}')

    return custom_ec2_dict


def get_ebs_vols(region, vol_status):
    volumes = []
    flat_volumes = []
    ebs_vol_list = []

    '''create paginator and build active ebs vols response dict'''

    print(f'{nl}Fetching EBS Volume Data from {region}...')
    pag = ec2_conn.get_paginator('describe_volumes')
    for page in pag.paginate(Filters=[
            {
                'Name': 'status',
                'Values': [vol_status]
            }
        ],
        PaginationConfig={
            'PageSize': 150
    }
    ):
        for i in page['Volumes']:
            volumes.append(i)
    for d in range(len(volumes)):
        flat_volumes.append(volumes[d])

    print(f'Found {len(flat_volumes)} EBS volumes in {region}')

    '''extract of all volumes IDs for subsequent snapshot query'''
    try:
        for i in range(len(flat_volumes)):
            ebs_vol_list.append(flat_volumes[i]['VolumeId'])
    except Exception as e:
        print(f'Failed to extract volume IDs:{nl}{e}')

    return ebs_vol_list, flat_volumes


def get_ebs_snapshots(region, ebs_vol_list):
    ebs_snaps = []
    '''fetch all snapshots associated with active EBS vols'''
    p = Paginator(ebs_vol_list, 150)
    for i in p:
        try:
            response = ec2_conn.describe_snapshots(Filters=[
                {
                    'Name': 'volume-id',
                    'Values': i,
                }
            ])
            for i in response['Snapshots']:
                ebs_snaps.append(i)

        except Exception as e:
            print(f'Error locating EBS Snaps from {region}: {e}')

    return ebs_snaps


def ebs_vols_report(region, vols_response, snap_list):
    custom_ebs_dict = []

    '''create snaptime dict for active snaps'''
    snaptime_dict = catalog_snaps(
        snap_list, 'VolumeId', 'StartTime')

    '''parse active ebs item level stats + dump to custom dictionary'''
    rbchck = None
    try:
        for i in range(len(vols_response)):
            snap_calendar = count_snaps(
                snaptime_dict, vols_response, i, 'VolumeId')

            '''encrypted true/false'''
            encryption = ''
            if vols_response[i]['Encrypted']:
                encryption = 'Yes'
            else:
                encryption = 'No'

            '''collect tag info'''
            try:
                tags = vols_response[i]['Tags']
            except Exception:
                tags = None
                pass

            '''determine true size of volume'''
            if not rbchck:
                print('\nCalculating consumed size of your In-Use EBS Vols')
                print('This may take up to 5 minutes per 100 Volumes\n')
                rbchck = True
            vol_id = vols_response[i]['VolumeId']
            consumed_blocks = vol_sizer(
                custom_ebs_dict, vol_id)

            custom_ebs_dict.append({
                'VolumeId': vols_response[i]['VolumeId'],
                'Size (GB)': vols_response[i]['Size'],
                'Consumed Size (GB)': consumed_blocks,
                'Region': str(region),
                'Disk Type': vols_response[i]['VolumeType'],
                'Encrypted?': encryption,
                'Snaps -24 Hours': snap_calendar['last_24'],
                'Snaps -7 Days': snap_calendar['last_7'],
                'Snaps -30 Days': snap_calendar['last_30'],
                'Snaps 30-365 Days': snap_calendar['_30to365'],
                'Snaps older than 1yr': snap_calendar['older'],
                'Total Snaps': snap_calendar['all_snaps'],
                'Tags': tags
            })

    except Exception as e:
        print(f'Failed to  write non-orphan ebs data to dict{nl}{e}')

    return custom_ebs_dict


def get_orphan_ebs_snaps(ebs_vol_list):
    response = []
    flat_response = []
    orphan_snaps = []

    print('\nFetching Orphan EBS Snapshot data...')

    pag = ec2_conn.get_paginator('describe_snapshots')
    for page in pag.paginate(OwnerIds=['self'],
                             PaginationConfig={'PageSize': 150}):
        for i in page['Snapshots']:
            response.append(i)

    '''flatten reponse structure'''
    for d in range(len(response)):
        flat_response.append(response[d])

    '''filter for orphans'''
    for s in flat_response:
        if s['VolumeId'] not in ebs_vol_list:
            orphan_snaps.append(s)

    return orphan_snaps


def ebs_snap_report(region, snap_response):
    ebs_snap_dict = []
    try:
        for i in range(len(snap_response)):
            '''check if snap is older than 60 days'''
            now = datetime.datetime.utcnow()
            snaptime = snap_response[i]['StartTime']
            snaptime = snaptime.replace(tzinfo=None)

            if snaptime <= (now - datetime.timedelta(days=60)):
                age = True
            else:
                age = False

            '''collect tag info'''
            try:
                tags = snap_response[i]['Tags']
            except Exception:
                tags = None
                pass

            '''break up date time to date + time of day'''
            snapdate = snaptime.strftime("%m/%d/%Y")
            snaphours = snaptime.strftime("%H:%M")

            '''calculate snapshot size'''
            snap_id = snap_response[i]['SnapshotId']
            vol_id = snap_response[i]['VolumeId']
            snap_size = snap_sizer(ebs_snap_dict, vol_id, snap_id)

            '''handle missing volume IDs (due to copied volumes)'''
            if snap_response[i]['VolumeId'] == 'vol-ffffffff':
                snap_size = None

            '''write out custom dictionary for orphan EBS snapshots'''
            ebs_snap_dict.append({
                'SnapshotId': snap_response[i]['SnapshotId'],
                'Source Vol ID': snap_response[i]['VolumeId'],
                'Source Vol Size (GB)': snap_response[i]['VolumeSize'],
                'Snapshot Size on S3 (GB)': snap_size,
                'Region': str(region),
                'Encrypted?': snap_response[i]['Encrypted'],
                'Snap Date': snapdate,
                'Snap Time': snaphours,
                '60+ Days Old': age,
                'Tags': tags
            })

    except Exception as e:
        print(f'{nl}Error locating Orphan EBS Snaps from {region}:{nl}{e}')
    return ebs_snap_dict


def analyze_engine_info(engine, engine_version, engine_mode):

    if engine_mode == "serverless":
        return False, False

    engine_version_parsed = engine_version.split(".")
    needs_upgrade = False

    try:
        if engine == "aurora" or engine == "aurora-mysql":
            major_version = (int(engine_version_parsed[0]), int(
                engine_version_parsed[1]))
            if major_version[0] != 5 or major_version[1] < 6:
                return False, False

            if major_version[1] == 6:
                needs_upgrade = engine_version < "5.6.mysql_aurora.1.19.2"
            elif major_version[1] == 7:
                needs_upgrade = engine_version < "5.7.mysql_aurora.2.04.4"

        elif engine == "mysql":
            major_version = (int(engine_version_parsed[0]), int(
                engine_version_parsed[1]))

            if major_version[0] == 5 and major_version[1] < 6:
                return False, False

            if major_version[1] == 6:
                needs_upgrade = engine_version < "5.6.40"
            elif major_version[1] == 7:
                needs_upgrade = engine_version < "5.7.24"
            elif major_version[1] == 8:
                needs_upgrade = engine_version < "8.0.13"

        elif engine == "postgres" or engine == "aurora-postgresql":
            major_version = (int(engine_version_parsed[0]), int(
                engine_version_parsed[1]))
            upgrade_targets = [((9, 4), 21), ((9, 5), 16),
                               ((9, 6), 12), ((10, 0), 7), ((11, 0), 2)]

            if major_version[0] > 11:
                return False, False

            elif major_version[0] == 9:
                for upgrade_target in upgrade_targets:
                    # ex: 9.6.12
                    if major_version[1] == upgrade_target[0][1] and int(engine_version_parsed[2]) < upgrade_target[1]:
                        needs_upgrade = True
                        break
            else:
                for upgrade_target in upgrade_targets:
                    # ex: 10.4
                    if major_version[0] == upgrade_target[0][0] and major_version[1] < upgrade_target[1]:
                        needs_upgrade = True
                        break
        else:
            return False, False
    except Exception:
        return False, False

    return True, needs_upgrade


def rds_inst_report(region):
    custom_rds_inst_dict = []
    flat_dbi_response = []
    db_instance_response = []
    rds_instance_ids = []
    rds_db_snaps = []
    describe_db_snaps_response = []

    '''fetch RDS instances with paginator'''
    print(f'{nl}Fetching RDS instance Data from {region}...')
    pag = rds_conn.get_paginator('describe_db_instances')
    for page in pag.paginate(PaginationConfig={'PageSize': 99}):
        for i in page['DBInstances']:
            db_instance_response.append(i)

    '''flatten response + add to dictionary'''
    for d in range(len(db_instance_response)):
        flat_dbi_response.append(db_instance_response[d])

    print(f'{nl}Found {len(flat_dbi_response)} RDS Instances in {region}')
    try:
        '''create list of rds instance ids'''
        for r in range(len(flat_dbi_response)):
            if ((flat_dbi_response[r]['Engine'] == 'aurora-postgresql'
                 or flat_dbi_response[r]['Engine'] == 'aurora'
                 or flat_dbi_response[r]['Engine'] == 'aurora-mysql')):
                pass
            else:
                rds_instance_ids.append(
                    flat_dbi_response[r]['DBInstanceIdentifier'])

    except Exception as e:
        print(f'failed to generate list of aurora instances {e}')

    '''read in all rds instance snapshots'''
    try:
        for i in rds_instance_ids:
            describe_db_snaps_response = rds_conn.describe_db_snapshots(
                MaxRecords=100,
                DBInstanceIdentifier=i)['DBSnapshots']
            for d in describe_db_snaps_response:
                rds_db_snaps.append(d)

    except Exception as e:
        print(f'{nl}Error locating RDS instance Snaps from {region}:{nl}{e}')

    '''parse snapshot times into dictionary'''
    snaptime_dict = catalog_snaps(
        rds_db_snaps, 'DBInstanceIdentifier', 'SnapshotCreateTime')

    '''define addtional variables + write selected data to RDS
    instance dictionary'''
    for i in range(len(flat_dbi_response)):
        snap_calendar = count_snaps(
            snaptime_dict, flat_dbi_response, i,
            'DBInstanceIdentifier')

        dbsize = None
        engine = flat_dbi_response[i]['Engine']
        engine_version = flat_dbi_response[i]['EngineVersion']

        '''dougs version checking algo'''
        is_granular_backup_supported, needs_internal_upgrade \
            = analyze_engine_info(engine, engine_version, "")

        try:
            if (engine == 'aurora-postgresql' or
               engine == 'aurora' or
               engine == 'aurora-mysql'):
                dbsize = None
            else:
                dbsize = flat_dbi_response[i]['AllocatedStorage']

        except Exception as e:
            print('''\nError retreiving aurora snapshot for ''' +
                  str(flat_dbi_response[i]['DBInstanceIdentifier']) +
                  '\n' + str(e))

        '''perform aurora name translations'''
        if engine == 'aurora':
            engine = 'aurora-mysql'

        '''lookup DB instance tags'''
        tags = list_rds_tags(
            flat_dbi_response[i]['DBInstanceArn'])

        custom_rds_inst_dict.append({
            'DBI Identifier': flat_dbi_response[i]['DBInstanceIdentifier'],
            'DB Provisioned Capacity (GB)': dbsize,
            'Region': str(region),
            'Status': flat_dbi_response[i]['DBInstanceStatus'],
            'Retention Period': flat_dbi_response[i]['BackupRetentionPeriod'],
            'Engine': engine,
            'EngineVersion': engine_version,
            'IsGranularBackupSupported': str(is_granular_backup_supported),
            'NeedsInternalUpgrade': str(needs_internal_upgrade),
            'DBInstanceStatus': flat_dbi_response[i]['DBInstanceStatus'],
            'Instance Type': flat_dbi_response[i]['DBInstanceClass'],
            'Snaps -24 Hours': snap_calendar['last_24'],
            'Snaps -7 Days': snap_calendar['last_7'],
            'Snaps -30 Days': snap_calendar['last_30'],
            'Snaps 30-365 Days': snap_calendar['_30to365'],
            'Snaps older than 1yr': snap_calendar['older'],
            'Total Snaps': snap_calendar['all_snaps'],
            'Tags': tags
        })

    return custom_rds_inst_dict


def rds_clusters_report(region):
    custom_rds_clusters_dict = []
    flat_dbc_response = []
    db_clusters_response = []
    rds_cluster_ids = []
    rds_cluster_snaps = []
    snap_calendar = []
    snaptime_dict = []

    '''create paginator and build rds_instance reponse dict'''
    print('\nFetching RDS cluster Data from ' + str(region) + '...')

    pag = rds_conn.get_paginator('describe_db_clusters')
    for page in pag.paginate(PaginationConfig={'PageSize': 100}):
        for i in page['DBClusters']:
            db_clusters_response.append(i)
    for d in range(len(db_clusters_response)):
        flat_dbc_response.append(db_clusters_response[d])

    print(f'{nl}Found {len(flat_dbc_response)} RDS Clusters in {region}')

    '''generate list of rds cluster IDs'''
    for i in range(len(flat_dbc_response)):
        rds_cluster_ids.append(
            flat_dbc_response[i]['DBClusterIdentifier'])

    '''read in all rds cluster snapshots'''
    try:
        for i in rds_cluster_ids:
            describe_db_snaps_response = get_rds_clustersnaps(i)
            for d in describe_db_snaps_response:
                rds_cluster_snaps.append(d)

    except Exception as e:
        print(f'{nl}Error locating rds instance Snaps from {region}: {e}')

    '''parse snapshot times into dictionary'''
    snaptime_dict = catalog_snaps(
        rds_cluster_snaps, 'DBClusterIdentifier', 'SnapshotCreateTime')

    '''write RDS data to data structure'''
    for i in range(len(flat_dbc_response)):
        snap_calendar = count_snaps(snaptime_dict, flat_dbc_response, i,
                                    'DBClusterIdentifier')

        allocatedGB = None
        engine = flat_dbc_response[i]['Engine']
        engine_version = flat_dbc_response[i]['EngineVersion']
        engine_mode = flat_dbc_response[i]['EngineMode']

        is_granular_backup_supported, needs_internal_upgrade \
            = analyze_engine_info(engine, engine_version, engine_mode)
        try:
            allocatedGB = get_aurora_allocated(
                flat_dbc_response[i]['DBClusterIdentifier'])
        except Exception as e:
            print(f'''failed to retreived DB size for
            {flat_dbc_response[i]['DBClusterIdentifier']}{nl} {e}''')

        '''perform aurora name translations'''
        if engine == 'aurora':
            engine = 'aurora-mysql'

        '''lookup db Tags'''
        tags = list_rds_tags(
            flat_dbc_response[i]['DBClusterArn'])

        custom_rds_clusters_dict.append({
            'DBC Identifier': flat_dbc_response[i]['DBClusterIdentifier'],
            'AllocatedStorage (GB)': allocatedGB,
            'Region': str(region),
            'Status': flat_dbc_response[i]['Status'],
            'Retention Period': flat_dbc_response[i]['BackupRetentionPeriod'],
            'Engine': engine,
            'EngineVersion': engine_version,
            'EngineMode': engine_mode,
            'IsGranularBackupSupported': str(is_granular_backup_supported),
            'NeedsInternalUpgrade': str(needs_internal_upgrade),
            'MultiAZ': flat_dbc_response[i]['MultiAZ'],
            'CrossAccountClone': flat_dbc_response[i]['CrossAccountClone'],
            'Snaps -24 Hours': snap_calendar['last_24'],
            'Snaps -7 Days': snap_calendar['last_7'],
            'Snaps -30 Days': snap_calendar['last_30'],
            'Snaps 30-365 Days': snap_calendar['_30to365'],
            'Snaps older than 1yr': snap_calendar['older'],
            'Total Snaps': snap_calendar['all_snaps'],
            'Tags': tags

        })

    return custom_rds_clusters_dict


def connect_service(key_id, secret_key, region, service):
    '''connect client'''
    print(f'{nl}Attmpting connection to {service} in {region}')
    try:
        client = boto3.client(
            service,
            aws_access_key_id=key_id,
            aws_secret_access_key=secret_key,
            region_name=region
        )
       
        return client
    except ClientError as e:
        print(f'{nl}Failed to connect to {service} in {region}:{nl}{e}')


def dataframer(dict, index_name, framename):
    try:
        df = pd.DataFrame(dict)
        df.set_index(index_name, inplace=True)
    except Exception as e:
        df = pd.DataFrame.from_records(
            [{index_name: 'No data', 'blank': 'No data'}], index=index_name)
        print(f'{nl}Error adding {framename} to dataframe.')
        print(f'Empty AWS response: {e}')
        print('''\nAre you sure you have the right permissions to
              access these AWS resources?''')

    return df


def get_rds_clustersnaps(clusterID):
    response = rds_conn.describe_db_cluster_snapshots(
        DBClusterIdentifier=clusterID,
        MaxRecords=100,
    )
    return response['DBClusterSnapshots']


def get_rds_instancesnaps(instanceID):
    response = rds_conn.describe_db_snapshots(
        DBInstanceIdentifier=instanceID,
        MaxRecords=100,
    )
    return response['DBSnapshots']


def get_aurora_allocated(clusterID):
    response = rds_conn.describe_db_cluster_snapshots(
        DBClusterIdentifier=clusterID,
        SnapshotType='automated',
        MaxRecords=100,
    )
    return response['DBClusterSnapshots'][0]['AllocatedStorage']


def list_rds_tags(arn):
    response = rds_conn.list_tags_for_resource(ResourceName=arn)
    return response['TagList']


def catalog_snaps(snap_list, target_id, timefield):
    '''
    need to feed in a list off all snapshots in a given region
    to match against a list of volumes or rds instances
    '''
    snaptime_dict = []
    try:
        '''loop for parseing snapshot times + adding to snaptime dict'''
        now = datetime.datetime.utcnow()
        for x in range(len(snap_list)):
            last_24 = 0
            last_7 = 0
            last_30 = 0
            _30to365 = 0
            older = 0

            yesterday = (now - datetime.timedelta(days=1))
            lastweek = (now - datetime.timedelta(days=7))
            lastmonth = (now - datetime.timedelta(days=30))
            lastyear = (now - datetime.timedelta(days=365))

            try:
                snaptime = snap_list[x][timefield]
                snaptime = snaptime.replace(tzinfo=None)
            except Exception:
                continue

            if snaptime > yesterday:
                last_24 += 1
            if (snaptime > lastweek) and (snaptime < yesterday):
                last_7 += 1
            if (snaptime > lastmonth) and (snaptime < lastweek):
                last_30 += 1
            if snaptime < lastmonth and (snaptime < lastyear):
                _30to365 += 1
            if snaptime < lastyear:
                older += 1

            snaptime_dict.append({
                target_id: snap_list[x][target_id],
                'last_24': last_24,
                'last_7': last_7,
                'last_30': last_30,
                '_30to365': _30to365,
                'older': older
            })

    except Exception as e:
        print('''Failed to parse snapshot time
              data in catalog fuction \n''' + str(e))

    return snaptime_dict


def count_snaps(snaptime_dict, describe_response, index, target_id):
    '''
    describe response = your boto response for describe_db_instances
    or describe_volumes etc

    target id = the index within your describe response with the item
    to match against ie volume-id or db_instance_id etc
    '''
    last_24 = 0
    last_7 = 0
    last_30 = 0
    _30to365 = 0
    older = 0
    response = []

    try:
        for s in range(len(snaptime_dict)):
            if snaptime_dict[s][target_id] == \
                    describe_response[index][target_id]:
                last_24 += snaptime_dict[s]['last_24']
                last_7 += snaptime_dict[s]['last_7']
                last_30 += snaptime_dict[s]['last_30']
                _30to365 += snaptime_dict[s]['_30to365']
                older += snaptime_dict[s]['older']

    except Exception as e:
        print('failed to count snaps' + str(e))

    all_snaps = last_24 + last_7 + last_30 + older + _30to365
    '''wrap variables into dict'''
    response = {
        'last_24': last_24,
        'last_7': last_7,
        'last_30': last_30,
        '_30to365': _30to365,
        'older': older,
        'all_snaps': all_snaps
    }

    return response


def vol_block_pager(snap_id):
    items = []
    response = ebs_conn.list_snapshot_blocks(SnapshotId=snap_id, MaxResults=5000)
    while True:
        if 'NextToken' in response:
            response = ebs_conn.list_snapshot_blocks(
                SnapshotId=snap_id,
                MaxResults=10000,
                NextToken=response['NextToken'])
            for i in response['Blocks']:
                items.append(i)
        if 'NextToken' not in response:
            return items
            break


def vol_sizer(custom_ebs_dict, vol_id):

    if read_blocks == 'true' or read_blocks == 'True':

        '''status updater'''
        length = len(custom_ebs_dict)
        if length % 10 == 0 and length != 0:
            print('''Still working. Parsed a total of ''' +
                  str(len(custom_ebs_dict)) + ' volumes')

        try:
            snap0_id = ec2_conn.describe_snapshots(Filters=[
                {
                    'Name': 'volume-id',
                    'Values': [vol_id],
                }
            ])['Snapshots'][0]['SnapshotId']
            blocks = vol_block_pager(snap0_id)
            consumed_blocks = int(512 * len(blocks) / 1024 / 1024)
        except Exception:
            consumed_blocks = None
    else:
        consumed_blocks = None

    return consumed_blocks


def snap_sizer(ebs_snap_dict, vol_id, snap_id):
    snap_gb_round = None
    
    if read_blocks == 'true' or read_blocks == 'True':
        '''status updater'''
        length = len(ebs_snap_dict)
        if length % 10 == 0 and length != 0:
            print('''Still working. Parsed a total of ''' +
                    str(len(ebs_snap_dict)) + ' snapshots')
        else:
            pass

        associated_snaps = ec2_conn.describe_snapshots(Filters=[
            {
                'Name': 'volume-id',
                'Values': [vol_id],
            }
        ])['Snapshots']

        for snap in range(len(associated_snaps)):
            if associated_snaps[snap] == associated_snaps[0]:
                blocks = vol_block_pager(snap_id)
                snap_gb = 512 * len(blocks) / 1024 / 1024
                snap_gb_round = int(snap_gb)
            else:
                snap0 = associated_snaps[(snap-1)]['SnapshotId']
                snap1 = snap_id
                blocks = ebs_changed_block_pager(snap0, snap1)
                snap_gb = 512 * len(blocks) / 1024 / 1024
                snap_gb_round = int(snap_gb)
    else:
        snap_gb_round = None

    return snap_gb_round


def ebs_changed_block_pager(snap0, snap1):
    items = []
    response = ebs_conn.list_changed_blocks(
        FirstSnapshotId=snap0, SecondSnapshotId=snap1, MaxResults=5000)
    for i in response['ChangedBlocks']:
        items.append(i)
    while True:
        if 'NextToken' in response:
            response = ebs_conn.list_changed_blocks(
                FirstSnapshotId=snap0,
                SecondSnapshotId=snap1,
                MaxResults=10000,
                NextToken=response['NextToken'])
            for i in response['ChangedBlocks']:
                items.append(i)
        if 'NextToken' not in response:
            return items
            break


class Paginator:
    def __init__(self, iterable, page_len=3):
        self.iterable = iterable
        self.page_len = page_len

    def __iter__(self):
        page = []
        for i in self.iterable:
            page.append(i)
            if len(page) == self.page_len:
                yield page
                page = []
        if page:
            yield page


if __name__ == '__main__':
    cli()
